#!/usr/bin/env python3
# encoding: utf-8

"""
server.py
server side responsible for the managing clients and scan execution flow.
"""
import hashlib
import hmac
import ssl
import os
import socket
import struct
import threading
from socketserver import TCPServer
from socketserver import ThreadingMixIn
from socketserver import BaseRequestHandler

from dscan.models.scanner import Context
from dscan.models.structures import Auth, Status
from dscan.models.structures import Command
from dscan.models.structures import Structure
from dscan import log


class DScanServer(ThreadingMixIn, TCPServer):
    """
    Foundation of the Server.
    Implements the shutdown with `threading.Event`, and ssl configurations.
    and injects `threading.Event` to the `RequestHandlerClass`
    """
    allow_reuse_address = True
    daemon_threads = True

    def __init__(self, *args, options, **kwargs):
        self._terminate = threading.Event()
        self.options = options
        self.ctx = Context.create(options)
        super().__init__(*args, **kwargs)

    @property
    def secret_key(self):
        """
        :return: str secret key generated in the config based on the
            certificate
        """
        return self.options.secret_key

    def get_request(self) -> tuple:
        """
        Used to add ssl support.

        :return: returns a `ssl.wrap_socket`
        :rtype: ´ssl.wrap_socket´
        """
        # noinspection PyTupleAssignmentBalance
        client, addr = super().get_request()
        # TODO: protocol version is hardcoded!
        client_ssl = ssl.wrap_socket(client, keyfile=self.options.sslkey,
                                     certfile=self.options.sslcert,
                                     ssl_version=ssl.PROTOCOL_TLSv1_2,
                                     ca_certs=None,
                                     server_side=True,
                                     do_handshake_on_connect=True,
                                     suppress_ragged_eofs=True,
                                     ciphers=self.options.ciphers)
        return client_ssl, addr

    def finish_request(self, request, client_address) -> BaseRequestHandler:
        """
        Finish one request by instantiating RequestHandlerClass.

        :return: `RequestHandlerClass`
        :rtype: ´RequestHandlerClass´
        """
        return self.RequestHandlerClass(request, client_address,
                                        self, terminate_event=self._terminate,
                                        context=self.ctx)

    def shutdown(self):
        """
         An override to allow a local terminate event to be set!
        """
        self._terminate.set()
        super().shutdown()
        self.server_close()
        if not self.ctx.is_finished:
            self.options.save_context(self.ctx)


class AgentHandler(BaseRequestHandler):
    HEADER = "<B"
    """
    Created when an agent connects, holds all the agents available actions.
    Terminates when scan targets finishes or an agent disconnects.
    """
    def __init__(self, *args, terminate_event, context, **kwargs):
        self._terminate = terminate_event
        self.ctx = context
        self.msg = None
        self.authenticated = False
        self.connected = False
        super().__init__(*args, **kwargs)

    @property
    def agent(self):
        """
        string representation of a connection ip:port.

        :return: str format of agent name ip:port
        """
        return "{}:{}".format(*self.client_address)

    @property
    def is_connected(self):
        """
        Check if the client is still connected, the terminate event has not
        been set and all stages are finished or not.

        :return: True if the client has disconnected or
            the terminate event has been triggered, else False
        :rtype: `bool`
        """
        return self.connected and not self._terminate.is_set() \
            and not self.ctx.is_finished

    def dispatcher(self):
        """
        Command dispatcher all logic to decode and dispatch the call.
        """
        self.msg = Structure.create(self.request)
        if not self.msg:
            self.connected = False
            log.info("Disconnected!")
            # mark any running task as interrupted
            # so that other agent can take it later
            self.ctx.interrupted(self.agent)
            return

        command_name = f"do_{self.msg.op_code.name.lower()}"
        if not hasattr(self, command_name):
            self.send_status(Status.FAILED)  # invalid command
            return

        command = getattr(self, command_name)
        # the only command authorized for unauthenticated agents
        if command_name != "do_AUTH" and not self.authenticated:
            self.send_status(Status.UNAUTHORIZED)
            self.connected = False
            self.request.close()
            return
        # call the command !
        command()

    def handle(self):
        """
        First method to be called by `BaseRequestHandler`.
        responsible for initial call to authentication `do_auth`,
        and `dispatcher`, the connection is kept alive as long as agent is
        connected and their are targets to be delivered.
        """
        log.info(f"{self.client_address} connected!")
        self.connected = True
        try:
            while self.is_connected:
                try:
                    # start by requesting authentication
                    if not self.authenticated:
                        self.do_auth()

                    self.dispatcher()
                except (socket.timeout, ConnectionError) as e:
                    log.info(f"{self.client_address} Timeout - {e}")
                    self.connected = False
                    # mark any running task as interrupted
                    # so that other agent can take it later
                    self.ctx.interrupted(self.agent)

                # wait a bit, in case a shutdown was requested!
                self._terminate.wait(1.0)
        finally:
            if self.ctx.is_finished:
                log.info("All stages are finished sending terminate event.")
                self.server.shutdown()
            self.request.close()

    def do_auth(self):
        """
        Handles the agent's authentication.
        """
        log.info(f"{self.client_address} initialized Authentication")
        challenge = os.urandom(128)
        self.request.sendall(Auth(challenge).pack())
        # wait for the client response

        self.msg = Structure.create(self.request)
        if not self.msg:
            # something's wrong with the message!
            self.send_status(Status.FAILED)
            return

        hmac_hash = hmac.new(self.server.secret_key, challenge, 'sha512')
        digest = hmac_hash.hexdigest().encode("utf-8")

        if hmac.compare_digest(digest, self.msg.data):
            log.info("Authenticated Successfully")
            self.authenticated = True
            self.send_status(Status.SUCCESS)
        else:
            self.send_status(Status.UNAUTHORIZED)
            self.request.close()

    def do_ready(self):
        """
        After the authentication the agent notifies the server, that is
        ready to start scanning.
        This will handle the request and send a target to be scanned.
        """
        log.info("is Ready for targets")

        log.info(f"Agent is running with uid {self.msg.uid}")
        if self.msg.uid != 0:
            log.info("Waning! agent is not running as root "
                     "syn scans might abort not enough privileges!")

        target_data = self.ctx.pop(self.agent)
        if not target_data:
            log.info("Target is None no more targets")
            # send empty command and terminate!
            cmd = Command("", "")
            self.request.sendall(cmd.pack())
            self.connected = False
            return

        cmd = Command(*target_data)
        self.request.sendall(cmd.pack())
        status_bytes = self.request.recv(1)

        if len(status_bytes) == 0:
            self.connected = False
            log.info("Disconnected!")
            self.ctx.interrupted(self.agent)
            return

        status, = struct.unpack("<B", status_bytes)
        if status == Status.SUCCESS.value:
            log.info("Started scanning !")
            self.ctx.running(self.agent)
        else:
            log.error("Scan command returned Error")
            log.info("Server is Terminating connection!")
            self.connected = False
            self.ctx.interrupted(self.agent)

    def do_report(self):
        """
        When the scan the ends, the agent notifies the server that is ready
        to send the report.
        This method will handle the report transfer save the report in the
        reports directory and make the target as finished
        if the file hashes match.
        """
        log.info("Agent Reporting Complete Scan!")
        log.info(f"Filename {self.msg.filename} total file size "
                 f"{self.msg.filesize} file hash {self.msg.filehash}")

        file_size = self.msg.filesize
        nbytes = 0
        report = self.ctx.get_report(self.agent,
                                     self.msg.filename.decode("utf-8"))
        try:
            digest = hashlib.sha512()
            self.ctx.downloading(self.agent)
            while nbytes < file_size:
                data = self.request.recv(1024)
                report.write(data)
                digest.update(data)
                nbytes = nbytes + len(data)

            if not hmac.compare_digest(digest.hexdigest().encode("utf-8"),
                                       self.msg.filehash):
                log.error(f"Files are not equal! {digest.hexdigest()}")
                self.send_status(Status.FAILED)
            else:
                log.info("files are equal!")
                self.ctx.completed(self.agent)
                self.send_status(Status.SUCCESS)
        finally:
            if report:
                report.flush()
                report.close()

    def send_status(self, code):
        """
        Sends a status code to the server.

        :param code:
        :type code: `dscan.models.structures.Status`
        """
        log.info(f"Sending status code {code}")
        response = struct.pack("<B", code)
        self.request.sendall(response)



#!/usr/bin/env python3
# encoding: utf-8

"""
client.py
client side responsible for the managing clients and scan execution flow.
"""
import os
import hmac
import struct
import random
import threading
from socket import socket
from socket import AF_INET
from socket import SOCK_STREAM
from socket import timeout
import ssl
from dscan import log
from dscan.models.structures import Structure
from dscan.models.structures import Auth
from dscan.models.structures import Ready
from dscan.models.structures import Status
from dscan.models.scanner import ScanProcess
from string import ascii_uppercase


class Agent:
    """
    Agent client implementation.
    """
    def __init__(self, config):
        """
        :param config: `dscan.models.scanner.Config`
        instance with the runtime configurations.
        """
        self.connected = False
        self.config = config
        self.con_retries = 0
        srv_hostname = config.srv_hostname
        ssl_context = ssl.create_default_context()
        ssl_context.check_hostname = False
        ssl_context.load_verify_locations(self.config.sslcert)
        self.socket = ssl_context.wrap_socket(socket(AF_INET, SOCK_STREAM),
                                              server_side=False,
                                              server_hostname=srv_hostname)
        self._terminate = threading.Event()
        self.scan = ScanProcess(self.config.outdir)

    def is_connected(self):
        """
        Check if the agent is still connected and not yet finished.

        :return: `True` if the client has disconnected or the terminate event
            has been triggered, else False.
        :rtype: `bool`
        """
        return self.con_retries < 3 and not self._terminate.is_set()

    def start(self):
        """
        Start the client connects to the server and authenticates.

        :return: True if was able to connect and authentication was
            successful else False.
        :rtype: `bool`
        """
        self.con_retries = 0
        # while the connection retry is under 3 tries
        # everytime the connection is interrupted the client
        # tries to connect authenticates and requests a target!
        while self.is_connected():
            try:
                self.socket.connect((self.config.host, self.config.port))
                self.connected = True
                if not self.do_auth():
                    # return out if auth was not successfully
                    self.connected = False
                    return
                # reset the counter if connection was successful.
                self.con_retries = 0
                # if authentication was successful request a target to scan.
                self.do_ready()
            except (timeout, ConnectionError, ValueError) as e:
                self.con_retries += 1
                log.error(f"Connection Timeout - {e}")
                log.error(f"Attempt - {self.con_retries} "
                          f"to establish connection")
                self.connected = False
                self.socket.close()
            finally:
                self.socket.close()

    def shutdown(self):
        """
        Set the terminate event On, to shutdown the agent.
        """
        self._terminate.set()

    def do_auth(self):
        """
        Initiate the authentication.

        :return: True if the status code of the last operation is
            `dscan.models.structures.Status.SUCCESS` False otherwise.
        :rtype: `bool`
        """
        log.info("Initiating authentication")
        opr = Structure.create(self.socket)
        if not opr:
            # unable to get message
            return False

        hmac_hash = hmac.new(self.config.secret_key, opr.data,
                             'sha512')
        digest = hmac_hash.hexdigest().encode("utf-8")

        self.socket.sendall(Auth(digest).pack())
        status_result = self.__check_status()
        if status_result:
            return status_result

    def do_ready(self):
        """
        This is recursive method and is responsible for, notifying the server
        is ready to execute a new scan, launch the scan and save the report.
        Until the server returns no target to scan.
        """
        alias = "".join(random.choice(ascii_uppercase) for _ in range(6))
        while self.connected:
            log.info("Requesting target...")
            self.socket.sendall(Ready(os.getuid(), alias).pack())
            cmd = Structure.create(self.socket)
            if not cmd:
                # unable to get message
                log.info("Unable to receive command from server")
                return

            if cmd.target.decode("utf-8") == "":
                log.info("server Send empty target Terminating!")
                self.con_retries = 3
                return

            log.info(f"Launching scan on {cmd}")
            report = self.scan.run(cmd.target.decode("utf-8"),
                                   cmd.options.decode("utf-8"),
                                   self.send_status)
            if report:
                self.socket.sendall(report.pack())
                if self.__send_report(report):
                    log.info("Report Transfer was successful")
                else:
                    log.error("Report Transfer was unsuccessful")
            else:
                self.send_status(Status.FAILED.value)

    def send_status(self, status):
        """
        :param status: int of a valid `dscan.models.structures.Status`
        """
        self.socket.sendall(struct.pack("<B", status))

    def __check_status(self):
        """
        Receives the status code from the server, and check the code value,
        see `dscan.models.structures.Status` for other valid status values.
        :return: True if the status code of the last operation is
        `dscan.models.structures.Status.SUCCESS` False otherwise.
        :rtype: `bool`
        """
        op_size = struct.calcsize("<B")
        op_bytes = self.socket.recv(op_size)
        if len(op_bytes) == 0:
            log.info("disconnected!")
            self.connected = False
            return False

        status, = struct.unpack("<B", op_bytes)
        if status == Status.SUCCESS.value:
            log.info("Operation Successful ...")
            return True
        else:
            log.error("Operation unsuccessful disconnecting...")
            return False

    def __send_report(self, report, retry=0):
        """
        Transfer the report.
        :param report: report message.
        :type report: `dscan.models.structures.Report`
        :param retry: number of attempts
        :type retry: `int`
        :return: bool the result of last try.
        :rtype: `bool`
        """
        nbytes = 0
        with open(os.path.join(self.config.outdir,
                               report.filename.decode("utf-8")),
                  "rb") as rfile:
            while nbytes < report.filesize:
                data = rfile.read(1024)
                self.socket.sendall(data)
                nbytes = nbytes + len(data)
        result = self.__check_status()
        if not result and retry < 3:
            retry += 1
            log.info(f"Transfer  retry {retry}")
            return self.__send_report(report, retry)
        else:
            return result



#!/usr/bin/env python3
# encoding: utf-8

import logging
import os
import threading
import shutil
from subprocess import Popen
from subprocess import PIPE
from configparser import ConfigParser, ExtendedInterpolation
from datetime import datetime
from dscan.models.parsers import parse_args
from dscan.client import Agent
from dscan.models.scanner import Config
from dscan.server import DScanServer
from dscan.server import AgentHandler
from dscan import dataPath
from dscan.out import ContextDisplay

FORMAT = '%(asctime)s - %(levelname)s - %(message)s'


def create_config(options):
    cfg = ConfigParser(interpolation=ExtendedInterpolation())
    conf_path = os.path.join(dataPath, 'dscan.conf')
    if options.config:
        conf_path = os.path.join(options.name, options.config)
    data = open(conf_path)
    cfg.read_file(data)
    data.close()
    config = Config(cfg, options)

    handler = logging.FileHandler(os.path.join(args.name, f"drecon-{now}.log"))
    handler.setFormatter(logging.Formatter(FORMAT))
    log = logging.getLogger()
    log.addHandler(handler)
    log.setLevel(logging.DEBUG)
    return config


def create_server(options):
    settings = create_config(options)
    settings.target_optimization(options.targets)
    server = DScanServer((settings.host, settings.port),
                         AgentHandler, options=settings)

    server_thread = threading.Thread(target=server.serve_forever)
    # Exit the server thread when the main thread terminates
    server_thread.daemon = True
    try:
        server_thread.start()
        logging.info(f"Server loop running in thread:{server_thread.name}")
        out = ContextDisplay(server.ctx)
        out.show()
        server_thread.join()
    except (KeyboardInterrupt, Exception) as ex:
        logging.info("Forced shutdown was requested!")
        logging.info(f"{ex}")
        if server:
            server.shutdown()


def create_agent(options):
    settings = create_config(options)
    agent = Agent(settings)
    agent_thread = threading.Thread(target=agent.start)
    try:
        agent_thread.start()
        agent_thread.join()
    except (KeyboardInterrupt, Exception) as ex:
        logging.info("Forced shutdown was requested!")
        logging.info(f"{ex}")
        if agent_thread:
            agent.shutdown()


def setup_config(options):
    os.makedirs(options.name, exist_ok=True)
    shutil.copy(os.path.join(dataPath, "dscan.conf"),
                os.path.join(options.name, "dscan.conf"))
    shutil.copy(os.path.join(dataPath, "agent.conf"),
                os.path.join(options.name, "agent.conf"))

    subj = f"/C={options.c}/ST={options.st}/L={options.l}" \
           f"/O={options.o}/OU={options.ou}/" \
           f"CN={options.cn}/emailAddress={options.email}"

    ssl_args = ['req', '-newkey', 'rsa:2048', '-nodes', '-keyout',
                os.path.join(options.name, 'keyfile.key'), '-x509', '-days',
                '3650', '-out', os.path.join(options.name, 'certfile.crt'),
                '-subj', subj
                ]
    with Popen(["openssl", *ssl_args], stdout=PIPE) as proc:
        print(proc.stdout.read())


def main():
    if args.cmd == "agent":
        create_agent(args)
    elif args.cmd == "srv":
        create_server(args)
    else:
        setup_config(args)


if __name__ == "__main__":
    now = datetime.now().strftime("%b-%d-%Y-%H-%M")
    parser = parse_args()
    args = parser.parse_args()
    main()



#!/usr/bin/env python3
# encoding: utf-8

"""
structures.py
network elements of the scanner
"""
import struct
from enum import IntEnum
from dscan import log


class Status(IntEnum):
    """
    Code values for the response status.
    """
    SUCCESS = 0x00
    UNAUTHORIZED = 0x01
    FINISHED = 0x02
    FAILED = 0xFF


class Operations(IntEnum):
    """
    Commands available!
    """
    AUTH = 0x01
    READY = 0x02
    COMMAND = 0x03
    STATUS_REQ = 0x04
    STATUS_RESP = 0x05
    REPORT = 0x06


class Structure:
    __slots__ = ()
    op_code = None
    _format = None
    HEADER = "<B"
    """
    The object representation for the info exchanged between
    agents and servers.
    """
    def __init__(self, *args, sock=None):
        if not sock:
            self.setData(*args)
        else:
            self.unpack(sock)

    def setData(self, *args):
        fields = getattr(self, '__slots__', [])
        if len(args) != len(fields):
            raise TypeError("Expected {} args".format(len(fields)))

        for name, value in zip(fields, args):
            if isinstance(value, str):
                value = value.encode("ascii")
            setattr(self, name, value)

    def unpack(self, sock):
        """
        Unpacks a known command based on the predefined
        :param sock:
        :return: Instance of a message subclass
        """
        if self._format:
            if isinstance(self._format, str):
                struct_size = struct.calcsize(self._format)
                data = struct.unpack(self._format,
                                     sock.recv(struct_size))
                return self.setData(*data)
            else:
                size_fmt = self._format[0]
                sz_nbytes = struct.calcsize(size_fmt)
                sz_bytes = sock.recv(sz_nbytes)
                sz = struct.unpack(size_fmt, sz_bytes)

                data_fmt = self._format[1].format(*sz)
                dt_nbytes = struct.calcsize(data_fmt)
                dt_bytes = sock.recv(dt_nbytes)
                data = struct.unpack(data_fmt, dt_bytes)
                return self.setData(*data)

    def pack(self):
        """
        Returns a `struct.pack` string ready to be sent
        :return: `struct.pack`
        """
        fmt = ''
        byte_order = ''
        values = [getattr(self, name, '') for name in self.__slots__]
        if self._format:
            if isinstance(self._format, str):
                if self._format.startswith(('<', '>', '!', '@')):
                    byte_order = self._format[0]
                    fmt = self._format[1:]
                fmt = f"{byte_order}B{fmt}"
                return struct.pack(fmt, self.op_code.value, *values)
            else:
                fmt = "{0}B{1}".format(*self._format)
                lengths = [len(getattr(self, name, '')) for name in
                           self.__slots__
                           if isinstance(getattr(self, name, ''), bytes)]
                fmt = fmt.format(*lengths)
                return struct.pack(fmt, self.op_code.value, *lengths, *values)

    @classmethod
    def create(cls, sock):
        try:
            op_size = struct.calcsize(cls.HEADER)
            op_bytes = sock.recv(op_size)
            if len(op_bytes) == 0:
                # agent disconnected !
                return
            op, = struct.unpack(cls.HEADER, op_bytes)

            subs = cls.__subclasses__()
            for operation in subs:
                if operation.op_code.value == op:
                    return operation(sock=sock)
            return None
        except (struct.error, ValueError) as e:
            log.info("Error parsing the message %s" % e)
            return None


class Auth(Structure):
    """
    Authentication Request
    from an Agent to server!
    """
    __slots__ = ('data', )
    _format = "<128s"

    op_code = Operations.AUTH

    def __str__(self):
        return f"Auth(op_code={self.op_code}, data={self.data})"


class Ready(Structure):
    """
    Ready to start Scan !
    Sent by an Agent to the server
    With the client's current user id
    """
    __slots__ = ('uid', 'alias')
    _format = ('<B', 'I{0}s')
    op_code = Operations.READY

    def __str__(self):
        return f"Ready({self.op_code}, uid={self.uid}, alias={self.alias})"


class Command(Structure):
    """
    Scan task information !
    Send by the server to the agent.
    final format is  <BB?s?s
    """
    __slots__ = ('target', "options")
    _format = ('<BB', '{0}s{1}s')
    op_code = Operations.COMMAND

    def __str__(self):
        return f"Command(op_code={self.op_code}, target={self.target}, " \
               f"options={self.options})"


class Report(Structure):
    """
    When an agent's task terminates,
    it initiates a report transfer request
    """
    __slots__ = ('filesize', 'filename', 'filehash')
    _format = ('<BB', 'I{0}s{1}s')
    op_code = Operations.REPORT

    def __str__(self):
        return f"Report(op_code={self.op_code}, filesize={self.filesize}," \
               f" filename={self.filename!s}, filehash={self.filehash})"
